import torch
from typing import Optional, Callable, Dict, Union, Any

ACT = Optional[type(torch.nn.Module)]   # activation type
INI = Optional[Callable]                # initializer type
TNS = torch.Tensor                      # Tensor
DTNS = Dict[str,Union[TNS,Any]]         # dict {str: Tensor or Any}


import numpy as np


# computes average of max and min probabilities for given array (batch) of probabilities
def avg_mm_probs(probs: np.ndarray) -> dict:
    probs = np.reshape(probs, (-1,probs.shape[-1]))
    max_probs = np.max(probs, axis=-1)  # max action_probs
    min_probs = np.min(probs, axis=-1)  # min action_probs
    return {
        'avg_max_prob':    np.mean(max_probs), # average of batch max action_prob
        'avg_min_prob':    np.mean(min_probs)} # average of batch min action_prob


import numpy as np
from typing import Optional, List


# processes zeroes array returned by model in following intervals
class ZeroesProcessor:

    def __init__(
            self,
            intervals: tuple=   (50,500,5000),
            tag_pfx=            'nane',     # prefix of tag in TB, (Not Activated NEurons)
            tbwr: Optional=     None):      # if given will put summaries to TB with intervals frequencies
        self.intervals = intervals
        self.zsL = {k: [] for k in self.intervals}
        self.single = []
        self.tag_pfx = tag_pfx
        self.tbwr = tbwr
        self.step = 0

    # takes next zeroes array and processes
    def process(self, zs:List[np.ndarray], step:Optional[int]=None):

        if step is None: step = self.step

        zs = np.concatenate(zs)
        self.single.append(np.mean(zs))

        rd = {}
        if len(self.single) == self.intervals[0]:
            rd[1] = np.mean(self.single)
            self.single = []

        for k in self.zsL:
            self.zsL[k].append(zs)
            if len(self.zsL[k]) == k:
                stacked = np.stack(self.zsL[k], axis=0) # joins arrays along 0 axis
                mean = np.mean(stacked, axis=0)         # mean along 0 axis (averages non activated over k)
                clipped = np.where(mean==1,1,0)         # where average over k is 1 leave 1, else 0
                rd[k] = np.mean(clipped)                # factor of neurons not activated over k (with 1)
                self.zsL[k] = []                        # reset

        if self.tbwr:
            for k in rd:
                self.tbwr.add(
                    value=  rd[k],
                    tag=    f'{self.tag_pfx}/nane_{k}',
                    step=   step)

        self.step += 1

        return rd


from pypaq.lipytools.pylogger import get_pylogger
import torch
from typing import List, Optional

from torchness.motorch import Module
from torchness.types import TNS, DTNS, INI
from torchness.base_elements import my_initializer
from torchness.layers import LayDense


# Simple Feats Classification Module
class SFeatsCSF(Module):

    def __init__(
            self,
            feats_width: int,
            in_drop: float=                         0.0,
            mid_width: int=                         30,
            mid_drop: float=                        0.0,
            num_classes: int=                       2,
            class_weights: Optional[List[float]]=   None,
            initializer: INI=                       None,
            dtype=                                  None,
            logger=                                 None):

        if not logger: logger = get_pylogger()
        self.logger = logger

        Module.__init__(self)

        if initializer is None: initializer = my_initializer

        self.drop = torch.nn.Dropout(p=in_drop) if in_drop else None

        self.logger.info(f'SFeatsCSF Module inits for feats of width {feats_width}')
        self.mid = LayDense(
            in_features=    feats_width,
            out_features=   mid_width,
            activation=     torch.nn.ReLU,
            bias=           True,
            initializer=    initializer,
            dtype=          dtype)

        self.mid_drop = torch.nn.Dropout(p=mid_drop) if mid_drop else None

        self.logits = LayDense(
            in_features=    mid_width,
            out_features=   num_classes,
            activation=     None,
            bias=           False,
            initializer=    initializer,
            dtype=          dtype)

        if class_weights:
            class_weights = torch.nn.Parameter(torch.tensor(class_weights), requires_grad=False)
        self.class_weights = class_weights

    def forward(self, feats:TNS) -> DTNS:
        if self.drop: feats = self.drop(feats)
        mid = self.mid(feats)
        if self.mid_drop: mid = self.mid_drop(mid)
        logits = self.logits(mid)
        probs = torch.nn.functional.softmax(logits, dim=-1)
        preds = torch.argmax(logits, dim=-1)
        return {
            'logits':   logits,
            'probs':    probs.detach().cpu().numpy(),
            'preds':    preds.detach().cpu().numpy()}

    def loss(self, feats:TNS, labels:TNS) -> DTNS:

        out = self.forward(feats)
        logits = out['logits']

        loss = torch.nn.functional.cross_entropy(
            input=      logits,
            target=     labels,
            weight=     self.class_weights,
            reduction=  'mean')
        acc = self.accuracy(logits, labels)
        f1 = self.f1(logits, labels)
        out.update({
            'loss': loss,
            'acc':  acc,
            'f1':   f1})
        return out


from pypaq.lipytools.pylogger import get_pylogger
from pypaq.lipytools.moving_average import MovAvg
import torch
from typing import Optional


# copied & refactored from torch.nn.utils.clip_grad.py, used by GradClipperAVT
def clip_grad_norm_(
        parameters,
        max_norm: float,
        norm_type: float=   2.0,
        do_clip: bool=      True, # disables clipping (just GN calculations)
) -> float:

    if isinstance(parameters, torch.Tensor): parameters = [parameters]
    parameters = [p for p in parameters if p.grad is not None]
    max_norm = float(max_norm)
    norm_type = float(norm_type)
    if len(parameters) == 0: return 0.0

    device = parameters[0].grad.device
    if norm_type == torch._six.inf:
        norms = [p.grad.detach().abs().max().to(device) for p in parameters]
        total_norm = norms[0] if len(norms) == 1 else torch.max(torch.stack(norms))
    else:
        total_norm = torch.norm(torch.stack([torch.norm(p.grad.detach(), norm_type).to(device) for p in parameters]), norm_type)

    if do_clip:
        clip_coef = max_norm / (total_norm + 1e-6)
        clip_coef_clamped = torch.clamp(clip_coef, max=1.0)
        for p in parameters:
            p.grad.detach().mul_(clip_coef_clamped.to(p.grad.device))

    return float(total_norm.cpu().numpy())

# gradient clipping class, clips gradients by clip_value or gg_avt_norm (computed with averaging window)
class GradClipperAVT:

    def __init__(
            self,
            module: torch.nn.Module,
            clip_value: Optional[float]=    None,   # clipping value, for None clips with avt
            avt_SVal: float=                0.1,    # start value for AVT (smaller value makes gradients warm-up)
            avt_window: int=                100,    # width of averaging window (number of steps)
            avt_max_upd: float=             1.5,    # max factor of gg_avt_norm to update with
            do_clip: bool=                  True,   # disables clipping (just GN calculations)
            logger=                         None):

        if not logger: logger = get_pylogger(name='ScaledLR')
        self._log = logger

        self.module = module
        self.clip_value = clip_value
        self.gg_avt_norm = avt_SVal
        self.avt_window = avt_window
        self.avt_max_upd = avt_max_upd
        self.do_clip = do_clip


    def clip(self):

        gg_norm = clip_grad_norm_(
            parameters= self.module.parameters(),
            max_norm=   self.clip_value or self.gg_avt_norm,
            do_clip=    self.do_clip)

        # in case of gg_norm explodes we want to update self.gg_avt_norm with value of self.avt_max_upd * self.gg_avt_norm
        avt_update = min(gg_norm, self.avt_max_upd * self.gg_avt_norm)
        self.gg_avt_norm = (self.gg_avt_norm * (self.avt_window-1) + avt_update) / self.avt_window # update
        self._log.debug(f'clipped with: gg_avt_norm({self.gg_avt_norm})')

        return {
            'gg_norm':      gg_norm,
            'gg_avt_norm':  self.gg_avt_norm}

# gradient clipping class, clips gradients by clip_value or gg_norm_mavg (computed with moving average)
class GradClipperMAVG:

    def __init__(
            self,
            module: torch.nn.Module,
            clip_value: Optional[float]=    None,   # clipping value, for None clips with mavg
            factor=                         0.01,
            first_avg=                      True,   # use averaging @start
            start_val=                      0.1,    # use this value @start (for first clip..)
            max_upd: float=                 1.5,    # max factor of gg_mavg to update with
            do_clip: bool=                  True):  # disables clipping (just GN calculations)

        self.module = module
        self.clip_value = clip_value

        self._gg_norm_mavg = MovAvg(factor=factor, first_avg=first_avg)
        self._gg_norm_mavg.upd(start_val)

        self.max_upd = max_upd
        self.do_clip = do_clip

    # clip & update parameters
    def clip(self):

        gg_norm = clip_grad_norm_(
            parameters= self.module.parameters(),
            max_norm=   self.clip_value or self._gg_norm_mavg(),
            do_clip=    self.do_clip)

        avt_update = min(gg_norm, self.max_upd * self._gg_norm_mavg()) # max value of update
        gg_norm_clip = self._gg_norm_mavg.upd(avt_update)

        return {
            'gg_norm':      gg_norm,
            'gg_norm_clip': gg_norm_clip}

    @property
    def gg_norm_clip(self):
        return self._gg_norm_mavg()